/* Package casbin provides middleware to enable ACL, RBAC, ABAC authorization support.

Simple example:

	package main

	import (
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/coreospackage/logginghelper"
		"github.com/casbin/casbin"
		"github.com/labstack/echo"
		"github.com/labstack/echo-contrib/casbin" casbin-mw
	)

	func main() {
		e := echo.New()

		// Mediate the access for every request
		e.Use(casbin-mw.Middleware(casbin.NewEnforcer("auth_model.conf", "auth_policy.csv")))

		e.Logger.Fatal(e.Start(":1323"))
	}

Advanced example:

	package main

	import (
		"github.com/casbin/casbin"
		"github.com/labstack/echo"
		"github.com/labstack/echo-contrib/casbin" casbin-mw
	)

	func main() {
		ce := casbin.NewEnforcer("auth_model.conf", "")
		ce.AddRoleForUser("alice", "admin")
		ce.AddPolicy(...)

		e := echo.New()

		echo.Use(casbin-mw.Middleware(ce))

		e.Logger.Fatal(e.Start(":1323"))
	}
*/

package casbin

import (
	"encoding/json"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/coreospackage/authhelper"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/coreospackage/logginghelper"
	"github.com/casbin/casbin"
	"github.com/labstack/echo"
	"github.com/labstack/echo/middleware"
)

type (
	// Config defines the config for CasbinAuth middleware.
	Config struct {
		// Skipper defines a function to skip middleware.
		Skipper middleware.Skipper

		// Enforcer CasbinAuth main rule.
		// Required.
		Enforcer *casbin.Enforcer
	}
)

var (
	// DefaultConfig is the default CasbinAuth middleware config.
	DefaultConfig = Config{
		Skipper: middleware.DefaultSkipper,
	}

	defaultJwtKey = "MKCLEMA-1234567890"
)

// Middleware returns a CasbinAuth middleware.
//
// For valid credentials it calls the next handler.
// For missing or invalid credentials, it sends "401 - Unauthorized" response.
func Middleware(ce *casbin.Enforcer, jwtKey string) echo.MiddlewareFunc {
	c := DefaultConfig
	c.Enforcer = ce
	defaultJwtKey = jwtKey
	return MiddlewareWithConfig(c)
}

// MiddlewareWithConfig returns a CasbinAuth middleware with config.
// See `Middleware()`.
func MiddlewareWithConfig(config Config) echo.MiddlewareFunc {
	// Defaults
	if config.Skipper == nil {
		config.Skipper = DefaultConfig.Skipper
	}

	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			if config.Skipper(c) || config.CheckPermission(c) {
				return next(c)
			}

			return echo.ErrForbidden
		}
	}
}

// GetUserName gets the user name from the request.
// Currently, only HTTP basic authentication is supported
func (a *Config) GetUserName(c echo.Context) string {
	username, _, _ := c.Request().BasicAuth()
	return username
}

// GetGroupNames gets the group names from the request.
func (a *Config) GetGroupNames(c echo.Context) []interface{} {
	var group []interface{}
	if c.Request().Header.Get("authorization") != "" {
		decodedToken, authError := authhelper.DecodeToken(c.Request().Header.Get("authorization"), defaultJwtKey)
		if authError != nil {
			logginghelper.LogError("error occured while decoding token ", authError)
			return nil
		}
		if decodedToken != nil {
			token := decodedToken["pre"]
			var dat interface{}
			_ = json.Unmarshal([]byte(token.(string)), &dat)

			if dat != nil && dat.(map[string]interface{})["group"] != nil {
				group = dat.(map[string]interface{})["group"].([]interface{})
			}
		}
	}

	return group
}

// CheckPermission checks the user/method/path combination from the request.
// Returns true (permission granted) or false (permission forbidden)
func (a *Config) CheckPermission(c echo.Context) bool {
	groups := a.GetGroupNames(c)
	method := c.Request().Method
	path := c.Request().URL.Path
	var i int
	var allowed = false
	for i = 0; i < len(groups); i++ {

		if a.Enforcer.Enforce(groups[i].(string), path, method) {
			allowed = true
			break
		}
	}
	return allowed
}
