package dalhelper

import (
	"sync"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/coreospackage/logginghelper"

	"github.com/noaway/heartbeat"
	"github.com/pquerna/ffjson/ffjson"

	"github.com/bluele/gcache"
)

// var PerformanceAnalyser = map[string]LazyCacheObject{}

// var isDebugMode = false
var lazyMutex = &sync.Mutex{}

const MIN_DATA_LENGTH = 10

// LazyFDPHelper Helps to Save and Get cache object
// also saves information into hard disk
type LazyFDPHelper struct {
	gc                     gcache.Cache
	Heartbeat              *heartbeat.Task
	IsProcessRunning       bool
	NumberOfUpdateAttempts int
	MAX_NUMBER_OF_RETRY    int
	RETRY_SLEEP_TIME_SEC   time.Duration
	INTERVAL_TIME_SEC      int
	MEMORY_WRITE_COUNT     int
	MEMORY_READ_COUNT      int
	DISK_READ_COUNT        int
	DISK_WRITE_COUNT       int
}

// LazyCacheObject is a
type LazyCacheObject struct {
	// Identifier will be used to identify the LazyCacheObject
	Identifier string
	// This will be used as a Key for GC cache
	FileName string

	// Data is for storing byte array of an object; right now it is not used.
	Data []byte

	// This will hold object which developer wants to fetch or store
	InterfaceData interface{}

	//This number indicates how many times InterfaceData data is changed
	ChangeCount int

	// This will tell if object is locked for file saving or not.
	IsLocked bool

	MEMORY_WRITE_COUNT int
	MEMORY_READ_COUNT  int
	DISK_READ_COUNT    int
	DISK_WRITE_COUNT   int
}

// StartProcess
func (lfd *LazyFDPHelper) StartProcess(objectCount int, taskName string,
	intervalTime int, sleepTime int, maxNumberOfRetry int, isDebugMode bool) {
	//This is Default value that is why used as
	lfd.INTERVAL_TIME_SEC = 5
	lfd.RETRY_SLEEP_TIME_SEC = 1
	// isDebugMode = isDebugMode
	// check parameter is valid - if not keep default else set new value
	if intervalTime > 0 {
		lfd.INTERVAL_TIME_SEC = intervalTime
	}

	// check parameter is valid - if not keep default else set new value

	if sleepTime > 0 {
		lfd.RETRY_SLEEP_TIME_SEC = time.Duration(sleepTime) * time.Millisecond
	}

	if maxNumberOfRetry <= 0 {
		lfd.MAX_NUMBER_OF_RETRY = 3
	} else {
		lfd.MAX_NUMBER_OF_RETRY = maxNumberOfRetry
	}

	//start Heartbeat event
	lfd.Heartbeat, _ = heartbeat.NewTast(taskName, lfd.INTERVAL_TIME_SEC)

	// Use default ARC algo for store
	lfd.gc = gcache.New(objectCount).
		LFU().
		Build()

	//Init timer
	lazyMutex.Lock()
	lfd.Heartbeat.Start(func() error {

		// check if process already running
		if lfd.IsProcessRunning {
			//If process is already running skip processing running it again.
			return nil
		}

		//process eatch object in cache and save to hdd.
		lfd.saveObjectsToFdb()
		return nil
	})
	lazyMutex.Unlock()
}

// saveObjectsToFdb is private method which will be called to save object on hdd
// this will fetch all objects in store and save it one by one
// right now it is not using any threading
func (lfd *LazyFDPHelper) saveObjectsToFdb() {
	lfd.IsProcessRunning = true

	// Fetch All Rows and then save into db
	for item := range lfd.gc.GetALL() {

		//TODO: catch errors
		cacheObjectraw, _ := lfd.gc.Get(item)
		cacheObjActual, _ := cacheObjectraw.(LazyCacheObject)

		if cacheObjActual.ChangeCount > 0 {

			cacheObjActual.IsLocked = true
			saveDataToFDB(cacheObjActual.FileName, cacheObjActual.InterfaceData)
			cacheObjActual.ChangeCount = 0
			cacheObjActual.IsLocked = false
			lfd.gc.Set(cacheObjActual.FileName, cacheObjActual)

			lazyMutex.Lock()
			lfd.MEMORY_WRITE_COUNT++
			lfd.DISK_WRITE_COUNT++

			// if isDebugMode {
			// 	lazyCacheObject := PerformanceAnalyser[cacheObjActual.FileName]
			// 	lazyCacheObject.MEMORY_WRITE_COUNT++
			// 	PerformanceAnalyser[cacheObjActual.FileName] = lazyCacheObject
			// }
			lazyMutex.Unlock()
		}
	}
	lfd.NumberOfUpdateAttempts = lfd.NumberOfUpdateAttempts + 1

	lfd.IsProcessRunning = false
}

// SaveOrUpdateDataInCache this method will Save object in cache if unavailable
func (lfd *LazyFDPHelper) SaveOrUpdateDataInCache(newObject LazyCacheObject) bool {

	byteArray, marshalError := ffjson.Marshal(newObject.InterfaceData)

	if marshalError != nil {
		logginghelper.LogError("error occured while marshaling data ", marshalError)
		return false
	}

	// check data length before saving. If less than assumed data length, return false

	lengthOfData := len(byteArray)
	if lengthOfData < MIN_DATA_LENGTH {
		logginghelper.LogError("data size is less than minimun expected data length. Actual data length: ", lengthOfData)
		logginghelper.LogError("data received: ", string(byteArray))

		return false
	}

	retryCount := 0
retrylabel:
	// Get prev object and then save new one
	//TODO: catch errors
	dataFromGC, getError := lfd.gc.Get(newObject.FileName)

	if getError != nil || dataFromGC == nil {
		newObject.ChangeCount = 1
		lfd.gc.Set(newObject.FileName, newObject)

	} else {

		oldObject, _ := dataFromGC.(LazyCacheObject)
		if oldObject.IsLocked && retryCount < lfd.MAX_NUMBER_OF_RETRY {
			retryCount++
			// Sleep for few sec so that other thread will release lock
			time.Sleep(lfd.RETRY_SLEEP_TIME_SEC)
			goto retrylabel
		}
		newObject.ChangeCount = oldObject.ChangeCount + 1
		lfd.gc.Set(newObject.FileName, newObject)
	}
	lazyMutex.Lock()
	lfd.MEMORY_WRITE_COUNT++
	lazyMutex.Unlock()
	// if isDebugMode {
	// 	lazyCacheObject := PerformanceAnalyser[newObject.FileName]
	// 	lazyCacheObject.MEMORY_WRITE_COUNT++
	// 	PerformanceAnalyser[newObject.FileName] = lazyCacheObject
	// }
	return true
}

//saveDataToFDB data to hard disk.
// This saves data by marshaling using json.Marshal
func saveDataToFDB(filePath string, objectData interface{}) bool {
	//TODO: catch errors
	byteArray, marshalError := ffjson.Marshal(objectData)
	if marshalError != nil {
		logginghelper.LogError("error occured while marshaling data ", marshalError)
		return false
	}
	// check data length before saving. If less than assumed data length, return false
	lengthOfData := len(byteArray)
	if lengthOfData < MIN_DATA_LENGTH {
		logginghelper.LogError("data size is less than minimun expected data length. Actual data length: ", lengthOfData)
		logginghelper.LogError("data received: ", string(byteArray))
		return false
	}

	saveError := SaveDataToFDB(filePath, byteArray, true)
	if saveError != nil {
		logginghelper.LogError("error occured while saving data ", saveError)
		return false
	}
	return true
}

// use it when req.
// func GetBytes(key interface{}) ([]byte, error) {
// 	var buf bytes.Buffer
// 	enc := gob.NewEncoder(&buf)
// 	err := enc.Encode(key)
// 	if err != nil {
// 		return nil, err
// 	}
// 	return buf.Bytes(), nil
// }

// RemoveDataFromCache Removes Data From Cache
func (lfd *LazyFDPHelper) RemoveDataFromCache(identifier string) {

	// Fetch All Rows and then save into db

	cachedObjectList := lfd.gc.GetALL()
	for item := range cachedObjectList {

		//TODO: catch errors
		cachedObject, getError := lfd.gc.Get(item)
		if getError != nil {
			logginghelper.LogError("error occured while getting ", item, " from gcache")
		}
		cachedObjectActual, _ := cachedObject.(LazyCacheObject)

		if cachedObjectActual.Identifier == identifier {
			if cachedObjectActual.ChangeCount > 0 {
				cachedObjectActual.IsLocked = true
				saveDataToFDB(cachedObjectActual.FileName, cachedObjectActual.InterfaceData)
				cachedObjectActual.ChangeCount = 0
				cachedObjectActual.IsLocked = false
				lazyMutex.Lock()
				lfd.DISK_WRITE_COUNT++
				lazyMutex.Unlock()
				// if isDebugMode {
				// 	lazyCacheObject := PerformanceAnalyser[cachedObjectActual.FileName]
				// 	lazyCacheObject.DISK_WRITE_COUNT++
				// 	PerformanceAnalyser[cachedObjectActual.FileName] = lazyCacheObject
				// }
			}
			lazyMutex.Lock()
			lfd.gc.Remove(cachedObjectActual.FileName)
			lazyMutex.Unlock()

		}
	}
}

//New Add new Key and value
// func  (lfd *LazyFDPHelper) SetToCache(newObject LazyCacheObject) error {
// 	lfd.MEMORY_WRITE_COUNT++
// 	return lfd.gc.Set(newObject.FileName, newObject.InterfaceData)
// }

//Get object based on key
func (lfd *LazyFDPHelper) GetFromCache(newObject LazyCacheObject) (interface{}, error) {
	lazyMutex.Lock()
	lfd.MEMORY_READ_COUNT++
	lazyMutex.Unlock()
	// if isDebugMode {
	// 	lazyCacheObject := PerformanceAnalyser[newObject.FileName]
	// 	lazyCacheObject.MEMORY_READ_COUNT++
	// 	PerformanceAnalyser[newObject.FileName] = lazyCacheObject
	// }
	return lfd.gc.Get(newObject.FileName)
}

// GetAll objects from gc
func (lfd *LazyFDPHelper) GetAllFromCache() map[interface{}]interface{} {
	return lfd.gc.GetALL()
}

// GetCacheLength Get Cache Length
func (lfd *LazyFDPHelper) GetCacheLength() int {
	return lfd.gc.Len()
}

// PurgeCache first saves all data inside FDB and finally purge all Cache
func (lfd *LazyFDPHelper) PurgeCache() {

	// Fetch All Rows and then save into db

	cachedObjectList := lfd.gc.GetALL()
	for item := range cachedObjectList {

		//TODO: catch errors
		cachedObject, getError := lfd.gc.Get(item)
		if getError != nil {
			logginghelper.LogError("error occured while getting ", item, " from gcache")
		}
		cachedObjectActual, conversionSuccessful := cachedObject.(LazyCacheObject)

		if conversionSuccessful && cachedObjectActual.ChangeCount > 0 {
			cachedObjectActual.IsLocked = true
			saveDataToFDB(cachedObjectActual.FileName, cachedObjectActual.InterfaceData)
			cachedObjectActual.ChangeCount = 0
			cachedObjectActual.IsLocked = false
			lazyMutex.Lock()
			lfd.DISK_WRITE_COUNT++
			lazyMutex.Unlock()
		}
	}
	lazyMutex.Lock()
	lfd.gc.Purge()
	lazyMutex.Unlock()
}

// =================================== > Lazywriter for appending data to file < ============================================

// StartProcessForAppend StartProcessForAppend
func (lfd *LazyFDPHelper) StartProcessForAppend(objectCount int, taskName string,
	intervalTime int, sleepTime int, maxNumberOfRetry int, isDebugMode bool) {
	//This is Default value that is why used as
	lfd.INTERVAL_TIME_SEC = 5
	lfd.RETRY_SLEEP_TIME_SEC = 1
	// isDebugMode = isDebugMode
	// check parameter is valid - if not keep default else set new value
	if intervalTime > 0 {
		lfd.INTERVAL_TIME_SEC = intervalTime
	}

	// check parameter is valid - if not keep default else set new value
	if sleepTime > 0 {
		lfd.RETRY_SLEEP_TIME_SEC = time.Duration(sleepTime) * time.Millisecond
	}

	if maxNumberOfRetry <= 0 {
		lfd.MAX_NUMBER_OF_RETRY = 3
	} else {
		lfd.MAX_NUMBER_OF_RETRY = maxNumberOfRetry
	}

	//start Heartbeat event
	lfd.Heartbeat, _ = heartbeat.NewTast(taskName, lfd.INTERVAL_TIME_SEC)

	// Use default ARC algo for store
	lfd.gc = gcache.New(objectCount).
		LFU().
		Build()

	//Init timer
	lazyMutex.Lock()
	lfd.Heartbeat.Start(func() error {

		// check if process already running
		if lfd.IsProcessRunning {
			//If process is already running skip processing running it again.
			return nil
		}

		//process eatch object in cache and save to hdd.
		lfd.appendObjectsToFdb()
		return nil
	})
	lazyMutex.Unlock()
}

// appendObjectsToFdb is private method which will be called to append object/data to file on hdd
// this will fetch all objects in store and append data to respective file it one by one
// right now it is not using any threading
func (lfd *LazyFDPHelper) appendObjectsToFdb() {
	lfd.IsProcessRunning = true

	// Fetch All Rows and then save into db
	for item := range lfd.gc.GetALL() {

		//TODO: catch errors
		cacheObjectraw, _ := lfd.gc.Get(item)
		cacheObjActual, _ := cacheObjectraw.(LazyCacheObject)

		if cacheObjActual.ChangeCount > 0 {

			cacheObjActual.IsLocked = true
			appendDataToFDB(cacheObjActual.FileName, cacheObjActual.InterfaceData)
			cacheObjActual.ChangeCount = 0
			cacheObjActual.IsLocked = false
			lfd.gc.Set(cacheObjActual.FileName, cacheObjActual)

			lazyMutex.Lock()
			lfd.MEMORY_WRITE_COUNT++
			lfd.DISK_WRITE_COUNT++

			// if isDebugMode {
			// 	lazyCacheObject := PerformanceAnalyser[cacheObjActual.FileName]
			// 	lazyCacheObject.MEMORY_WRITE_COUNT++
			// 	PerformanceAnalyser[cacheObjActual.FileName] = lazyCacheObject
			// }
			lazyMutex.Unlock()
		}
	}
	lfd.NumberOfUpdateAttempts = lfd.NumberOfUpdateAttempts + 1

	lfd.IsProcessRunning = false
}

// appendDataToFDB appends data to file on hard drive.
// This appends data by marshaling using json.Marshal
// '\n' will be added as a separator between two different objects
func appendDataToFDB(filePath string, objectData interface{}) {
	//TODO: catch errors
	byteArray, marshalError := ffjson.Marshal(objectData)
	if marshalError != nil {
		logginghelper.LogError("error occured while marshaling data ", marshalError)
	}
	byteArray = append(byteArray, []byte("\n")...)
	saveError := AppendDataToFDB(filePath, byteArray)
	if saveError != nil {
		logginghelper.LogError("error occured while saving data ", saveError)
	}
}

func Save(filePath string, data interface{}) {
	saveDataToFDB(filePath, data)
}

// RemoveDataFromCacheForAppend Removes Data From Cache
func (lfd *LazyFDPHelper) RemoveDataFromCacheForAppend(identifier string) {

	// Fetch All Rows and then save into db

	cachedObjectList := lfd.gc.GetALL()
	for item := range cachedObjectList {

		//TODO: catch errors
		cachedObject, getError := lfd.gc.Get(item)
		if getError != nil {
			logginghelper.LogError("error occured while getting ", item, " from gcache")
		}
		cachedObjectActual, _ := cachedObject.(LazyCacheObject)
		if cachedObjectActual.Identifier == identifier {
			if cachedObjectActual.ChangeCount > 0 {
				cachedObjectActual.IsLocked = true
				appendDataToFDB(cachedObjectActual.FileName, cachedObjectActual.InterfaceData)
				cachedObjectActual.ChangeCount = 0
				cachedObjectActual.IsLocked = false
				lazyMutex.Lock()
				lfd.DISK_WRITE_COUNT++
				lazyMutex.Unlock()
				// if isDebugMode {appendDataToFDB
				// 	lazyCacheObject := PerformanceAnalyser[cachedObjectActual.FileName]
				// 	lazyCacheObject.DISK_WRITE_COUNT++
				// 	PerformanceAnalyser[cachedObjectActual.FileName] = lazyCacheObject
				// }
			}
			lazyMutex.Lock()
			lfd.gc.Remove(cachedObjectActual.FileName)
			lazyMutex.Unlock()

		}
	}
}
