package downloadhelper

import (
	"errors"
	"os"
	"path"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/coreospackage/filehelper"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/coreospackage/securityhelper"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/coreospackage/logginghelper"
	"github.com/cavaliercoder/grab"
)

const (
	EMPTY_STR     = ""
	HTTP_PROTOCOL = "http"
	ERR_FILE_HASH = 6273839363813614546
	FILE_SIZE     = 37
)

//DownloadFile file from cloud
// sourcePath : sourcePath
// destinationPath : destinationPath
// jwtToken if any
// retryCnt : how many times process should retry for download.
func DownloadFile(sourcePath, destinationPath, jwtToken string, retryCnt int) (bool, error) {
	logginghelper.LogDebug("IN : DownloadFile")
	destinationPath = filehelper.CleanPath(destinationPath)
	mkclCacheBurst := strconv.FormatInt(time.Now().UnixNano(), 10)
	if strings.Contains(sourcePath, "?") {
		sourcePath = sourcePath + "&mkclCacheBurst=" + mkclCacheBurst
	} else {
		sourcePath = sourcePath + "?mkclCacheBurst=" + mkclCacheBurst
	}
	logginghelper.LogDebug("DownloadFile : Downloading ", sourcePath, " at ", destinationPath)
	if strings.Trim(sourcePath, EMPTY_STR) == EMPTY_STR || strings.Trim(destinationPath, EMPTY_STR) == EMPTY_STR || !strings.HasPrefix(sourcePath, HTTP_PROTOCOL) {
		logginghelper.LogDebug("DownloadFile : sourcePath is empty")
		logginghelper.LogDebug("OUT : DownloadFile")
		return false, errors.New("Either Source/Destination url is empty or source url does not start with 'http'")
	}
	if filehelper.FileAvailabilityCheck(destinationPath) {
		destInfo, infoError := filehelper.FileInfo(destinationPath)
		if infoError != nil {
			logginghelper.LogDebug("DownloadFile : Destination Path Information Error")
			logginghelper.LogDebug("OUT : DownloadFile")
			return false, errors.New("error occured while getting information of destination directory : " + destinationPath)
		}

		if destInfo.IsDir() {
			logginghelper.LogDebug("DownloadFile : Destination Path is Directory")
			logginghelper.LogDebug("OUT : DownloadFile")
			return false, errors.New("destination path must be a file path, got directory : " + destinationPath)
		}
	}
	client := grab.NewClient()
	// Changed because update was failing - Ajitj
	// client.HTTPClient.Timeout = time.Second * 10
	req, _ := grab.NewRequest(destinationPath, sourcePath)
	if jwtToken != "NA" {
		req.HTTPRequest.Header.Add("Authorization", jwtToken)
	}

	for index := 1; index <= retryCnt; index++ {
		err := BackupFile(destinationPath)
		if err != nil {
			logginghelper.LogError("GrabDownloadFile Backup Error : ", err)
			logginghelper.LogDebug("OUT : DownloadFile")
			return false, err
		}
		resp := client.Do(req)
		if err := resp.Err(); err != nil {
			logginghelper.LogError("Download failed: ", err)
			RestoreBackup(destinationPath)
			time.Sleep(200 * time.Millisecond)
			continue
		}

		if resp == nil || resp.HTTPResponse == nil {
			logginghelper.LogError("Download failed: ", err)
			RestoreBackup(destinationPath)
			time.Sleep(200 * time.Millisecond)
			continue
		} else if resp.HTTPResponse != nil && resp.HTTPResponse.StatusCode != 200 && resp.HTTPResponse.StatusCode != 206 {
			logginghelper.LogError("Download failed.")
			RestoreBackup(destinationPath)
			time.Sleep(200 * time.Millisecond)
			continue
		} else {
			fileInfo, err := filehelper.FileInfo(destinationPath)
			if err != nil {
				logginghelper.LogError("Download failed: ", err)
				RestoreBackup(destinationPath)
			}
			if fileInfo.Size() < FILE_SIZE {
				hash, err := securityhelper.GetHashChecksumOfFile(destinationPath)
				if err != nil {
					logginghelper.LogError("Download failed: ", err)
					RestoreBackup(destinationPath)
					time.Sleep(200 * time.Millisecond)
					continue
				}
				if hash == ERR_FILE_HASH {
					logginghelper.LogError("Download failed: ", err)
					RestoreBackup(destinationPath)
					time.Sleep(200 * time.Millisecond)
					continue
				}
			}
		}
		if resp.HTTPResponse != nil && (resp.HTTPResponse.StatusCode == 200 || resp.HTTPResponse.StatusCode == 206) {
			closeErr := resp.HTTPResponse.Body.Close()
			if closeErr != nil {
				logginghelper.LogError("Error while closing response : ", closeErr)
			}
			logginghelper.LogDebug("DownloadFile : Download Successful")

			timeStampChangeError := changeFileTimes(resp, destinationPath)
			if timeStampChangeError != nil {
				logginghelper.LogError("Error while changing timestamp for file : ", destinationPath, " : ", timeStampChangeError)
			}

			DeleteBackup(destinationPath)
			logginghelper.LogDebug("OUT : DownloadFile")
			return true, nil
		}
		logginghelper.LogError("Download failed: ", err)
		RestoreBackup(destinationPath)
		logginghelper.LogDebug("OUT : DownloadFile")
		return false, nil

	}
	logginghelper.LogError("Download failed.")
	logginghelper.LogDebug("OUT : DownloadFile")
	return false, errors.New("File not found on cloud")
}

// BackupFile Backup File
func BackupFile(filePath string) error {
	logginghelper.LogDebug("IN : BackupFile")
	fileExists := filehelper.FileAvailabilityCheck(filePath)
	if fileExists {
		logginghelper.LogDebug("BackupFile : File Exists. Creating Backup")
		parentDirectory, fileName := path.Split(filePath)
		ext := path.Ext(fileName)
		backupFileName := strings.Replace(fileName, ext, "", -1) + "_backup" + ext

		backupFilePath := filepath.Join(parentDirectory, backupFileName)
		logginghelper.LogDebug("OUT : BackupFile")
		return filehelper.RenameFile(filePath, backupFilePath)
	}
	logginghelper.LogDebug("BackupFile : File Not Exists.")
	logginghelper.LogDebug("OUT : BackupFile")
	return nil
}

// RestoreBackup Restores Backup
func RestoreBackup(filePath string) error {
	logginghelper.LogDebug("IN : RestoreBackup")
	fileExists := filehelper.FileAvailabilityCheck(filePath)
	if fileExists {
		logginghelper.LogDebug("RestoreBackup : File Exists. Deleting File")
		err := filehelper.DeleteFile(filePath)
		if err != nil {
			logginghelper.LogError("RestoreBackup File Delete Error: ", err)
			return err
		}
	}
	parentDirectory, fileName := path.Split(filePath)
	ext := path.Ext(fileName)
	backupFileName := strings.Replace(fileName, ext, "", -1) + "_backup" + ext

	backupFilePath := filepath.Join(parentDirectory, backupFileName)
	fileExists = filehelper.FileAvailabilityCheck(backupFilePath)
	if fileExists {
		logginghelper.LogDebug("RestoreBackup : File Exists. Restoring Backup")
		logginghelper.LogDebug("OUT : RestoreBackup")
		return filehelper.RenameFile(backupFilePath, filePath)
	}
	logginghelper.LogDebug("RestoreBackup : Backup File Not Exists.")
	logginghelper.LogDebug("OUT : RestoreBackup")
	return nil
}

// DeleteBackup Deletes Backup
func DeleteBackup(filePath string) error {
	logginghelper.LogDebug("IN : DeleteBackup")
	parentDirectory, fileName := path.Split(filePath)
	ext := path.Ext(fileName)
	backupFileName := strings.Replace(fileName, ext, "", -1) + "_backup" + ext

	backupFilePath := filepath.Join(parentDirectory, backupFileName)
	fileExists := filehelper.FileAvailabilityCheck(backupFilePath)
	if fileExists {
		logginghelper.LogDebug("DeleteBackup : File Exists. Deleting Backup")
		logginghelper.LogDebug("OUT : DeleteBackup")
		return filehelper.DeleteFile(backupFilePath)
	}
	logginghelper.LogDebug("DeleteBackup : Backup File Not Exists.")
	logginghelper.LogDebug("OUT : DeleteBackup")
	return nil
}

func changeFileTimes(resp *grab.Response, destination string) error {
	logginghelper.LogDebug("IN : ChangeFileTimes")

	timeHeader := resp.HTTPResponse.Header.Get("Last-Modified")
	logginghelper.LogInfo("Header Time:", timeHeader)

	timeUTC, parseError := time.Parse("Mon, 02 Jan 2006 15:04:05 GMT", timeHeader)
	if parseError != nil {
		logginghelper.LogError(parseError)
		return parseError
	}
	localtime := timeUTC.Local()
	logginghelper.LogInfo("Local Time:", localtime)
	changeError := os.Chtimes(destination, localtime, localtime)
	if changeError != nil {
		logginghelper.LogError(changeError)
		return changeError
	}
	logginghelper.LogDebug("OUT : ChangeFileTimes")

	return nil
}
