package filehelper

import (
	"archive/zip"
	"io"
	"io/ioutil"
	"os"
	"path"
	"path/filepath"
	"strings"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/coreospackage/logginghelper"
	// "github.com/klauspost/compress/zip" because this do not work from windows to linux
)

//FileHelperServiceObject FileHelperServiceObject must be created while calling FileSearh function
type FileHelperServiceObject struct {
	searchResult   []string
	searchFileName string
}

// ReadFile reads contents from provided file path
func ReadFile(filePath string) ([]byte, error) {
	return ioutil.ReadFile(filePath)

}

//CreateFile creates a new file
func CreateFile(filePath string) (*os.File, error) {
	return os.Create(filePath)
}

// WriteFile writes provided  bytes to file
func WriteFile(filePath string, data []byte) error {
	return ioutil.WriteFile(filePath, data, 0644)
}

//AppendFile appends prvided data/text to file
func AppendFile(filename string, text string) (int, error) {

	f, err := os.OpenFile(filename, os.O_CREATE|os.O_APPEND|os.O_WRONLY, 0600)
	if err != nil {
		return 0, err
	}

	defer f.Close()

	return f.WriteString(text)
}

//DeleteFile deletes provided file path
func DeleteFile(filePath string) error {
	return os.Remove(filePath)
}

// RenameFile renames/moves file from old path to new path
func RenameFile(oldFilePath, newFilePath string) error {
	return os.Rename(oldFilePath, newFilePath)
}

// CreateDirectory creates directory using provided path
func CreateDirectory(directoryPath string) error {
	return os.Mkdir(directoryPath, os.ModePerm)
}

// CreateDirectoryRecursive creates directory recursively using provided path
func CreateDirectoryRecursive(directoryPath string) error {
	return os.MkdirAll(directoryPath, os.ModePerm)
}

// DeleteDirectory creates directory using provided path
func DeleteDirectory(directoryPath string) error {
	return os.RemoveAll(directoryPath)
}

//ListDirectory returns list of all available components of directory
func ListDirectory(directoryPath string) ([]os.FileInfo, error) {
	return ioutil.ReadDir(directoryPath)
}

//Zip Zip
//FIXME: Use only one type of zip like zstd
func Zip(source, target string) error {
	source = filepath.Clean(source)
	target = filepath.Clean(target)

	zipfile, err := os.Create(target)
	if err != nil {
		return err
	}
	defer zipfile.Close()

	archive := zip.NewWriter(zipfile)
	defer archive.Close()

	info, err := os.Stat(source)
	if err != nil {
		return err
	}

	var baseDir string
	if info.IsDir() {
		baseDir = filepath.Base(source)
	}

	filepath.Walk(source, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			return err
		}

		header, err := zip.FileInfoHeader(info)
		if err != nil {
			return err
		}

		if baseDir != "" {
			header.Name = filepath.Join(baseDir, strings.TrimPrefix(path, source))
			// Replace all occurances of \\ with /. This is necessary to properly unzip the zip file created on windows system on ubuntu system
			header.Name = strings.Replace(header.Name, "\\", "/", -1)
		}

		if info.IsDir() {
			header.Name += "/"
		} else {
			header.Method = zip.Deflate
		}

		writer, err := archive.CreateHeader(header)
		if err != nil {
			return err
		}

		if info.IsDir() {
			return nil
		}

		file, err := os.Open(path)
		if err != nil {
			return err
		}
		_, err = io.Copy(writer, file)
		if err != nil {
		}

		file.Close()
		return err
	})

	return err
}

//ZipWithoutBaseDirectory Zip Without Base Directory
func ZipWithoutBaseDirectory(source, target string) error {
	source = filepath.Clean(source)
	target = filepath.Clean(target)

	zipfile, err := os.Create(target)
	if err != nil {
		return err
	}
	defer zipfile.Close()

	archive := zip.NewWriter(zipfile)
	defer archive.Close()

	info, err := os.Stat(source)
	if err != nil {
		return err
	}

	var baseDir string
	if info.IsDir() {
		baseDir = filepath.Base(source)
	}

	filepath.Walk(source, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			return err
		}

		header, err := zip.FileInfoHeader(info)
		if err != nil {
			return err
		}

		if baseDir != "" {
			header.Name = strings.TrimPrefix(path, source)
			// Replace all occurances of \\ with /. This is necessary to properly unzip the zip file created on windows system on ubuntu system
			header.Name = strings.Replace(header.Name, "\\", "/", -1)
		}

		if info.IsDir() {
			header.Name += "/"
		} else {
			header.Method = zip.Deflate
		}

		writer, err := archive.CreateHeader(header)
		if err != nil {
			return err
		}

		if info.IsDir() {
			return nil
		}

		file, err := os.Open(path)
		if err != nil {
			return err
		}
		_, err = io.Copy(writer, file)
		if err != nil {
		}

		file.Close()
		return err
	})

	return err
}

//ZipWithSkipFileList This method will skip the file added in skiplist from zip
func ZipWithSkipFileList(source, target string, skipFileList []string) error {

	source = filepath.Clean(source)
	target = filepath.Clean(target)
	zipfile, err := os.Create(target)
	if err != nil {
		return err
	}
	defer zipfile.Close()
	archive := zip.NewWriter(zipfile)
	defer archive.Close()
	info, err := os.Stat(source)
	if err != nil {
		return err
	}
	var baseDir string
	if info.IsDir() {
		baseDir = filepath.Base(source)
	}
	filepath.Walk(source, func(path string, info os.FileInfo, err error) error {

		if !CheckPathInSkipList(info.Name(), skipFileList) {
			header, err := zip.FileInfoHeader(info)
			if err != nil {
				return err
			}

			if baseDir != "" {
				header.Name = filepath.Join(baseDir, strings.TrimPrefix(path, source))
				// Replace all occurances of \\ with /. This is necessary to properly unzip the zip file created on windows system on ubuntu system
				header.Name = strings.Replace(header.Name, "\\", "/", -1)
			}

			if info.IsDir() {
				header.Name += "/"
			} else {
				header.Method = zip.Deflate
			}

			writer, err := archive.CreateHeader(header)
			if err != nil {
				return err
			}

			if info.IsDir() {
				return nil
			}

			file, err := os.Open(path)
			if err != nil {
				return err
			}

			_, err = io.Copy(writer, file)
			if err != nil {
			}

			file.Close()
			return nil
		}

		return nil
	})
	return err
}

func CheckPathInSkipList(path string, skipListForActivation []string) bool {
	for _, prefix := range skipListForActivation {
		if strings.HasPrefix(path, prefix) {
			return true
		}
		if strings.HasSuffix(path, prefix) {
			return true
		}
	}
	return false
}

//Unzip Unzip
func Unzip(archive, target string) error {
	reader, err := zip.OpenReader(archive)
	if err != nil {
		logginghelper.LogError(err)
		return err
	}

	mkcDirError := os.MkdirAll(target, 0755)
	if mkcDirError != nil {
		logginghelper.LogError(mkcDirError)
		return mkcDirError
	}

	for _, file := range reader.File {
		filePath := filepath.Join(target, file.Name)
		parentPath := path.Dir(CleanPath(filePath))

		if !FileAvailabilityCheck(parentPath) {
			CreateDirectoryRecursive(parentPath)
		}

		if file.FileInfo().IsDir() {
			os.MkdirAll(filePath, file.Mode())
			continue
		}

		fileReader, openError := file.Open()
		if openError != nil {
			logginghelper.LogError(openError)
			return openError
		}

		targetFile, targetOpenError := os.OpenFile(filePath, os.O_WRONLY|os.O_CREATE|os.O_TRUNC, file.Mode())
		if targetOpenError != nil {
			logginghelper.LogError(targetOpenError)
			return targetOpenError
		}

		_, copyError := io.Copy(targetFile, fileReader)
		if copyError != nil {
			logginghelper.LogError(copyError)
			return copyError
		}
		targetCloseError := targetFile.Close()
		if targetCloseError != nil {
			logginghelper.LogError(targetCloseError)
			return targetCloseError
		}

		fileCloseError := fileReader.Close()
		if fileCloseError != nil {
			logginghelper.LogError(fileCloseError)
			return fileCloseError
		}
	}
	closeError := reader.Close()
	if closeError != nil {
		logginghelper.LogError(closeError)
		return closeError
	}
	return nil
}

//MoveDirectory MoveDirectory
func MoveDirectory(source, destination string) error {
	return os.Rename(source, destination)
}

//MoveFile MoveFile
func MoveFile(source, destination string) error {
	return os.Rename(source, destination)
}

// MoveFileToOtherHost MoveFileToOtherHost
func MoveFileToOtherHost(source, destination string) error {
	// Copy source file to destination
	copyError := CopyFile(source, destination)
	if copyError != nil {
		logginghelper.LogError("error occured while copying file : ", copyError)
		return copyError
	}

	// Delete source file
	deleteError := DeleteFile(source)
	if deleteError != nil {
		logginghelper.LogError("error occured while deleting file : ", deleteError)
		return deleteError
	}
	return nil
}

//CopyFile CopyFile
func CopyFile(source, destination string) error {
	// return os.Rename(source, destination)

	data, err := ioutil.ReadFile(source)
	if err != nil {
		return err
	}
	err = ioutil.WriteFile(destination, data, 0644)
	if err != nil {
	}
	return err
}

//CopyDir makes copy of source directory to the destination directory
func CopyDir(source string, dest string) (err error) {

	// get properties of source dir
	sourceinfo, err := os.Stat(source)
	if err != nil {
		return err
	}

	// create dest dir

	err = os.MkdirAll(dest, sourceinfo.Mode())
	if err != nil {
		return err
	}

	directory, _ := os.Open(source)

	objects, err := directory.Readdir(-1)

	for _, obj := range objects {

		sourcefilepointer := source + "/" + obj.Name()

		destinationfilepointer := dest + "/" + obj.Name()

		if obj.IsDir() {
			// create sub-directories - recursively
			err = CopyDir(sourcefilepointer, destinationfilepointer)
			if err != nil {
				logginghelper.LogError(err)
			}
		} else {
			// perform copy
			err = CopyFile(sourcefilepointer, destinationfilepointer)
			if err != nil {
				logginghelper.LogError(err)
			}
		}

	}
	return
}

//ReplaceFile ReplaceFile
func ReplaceFile(data []byte, destination string) error {

	err := ioutil.WriteFile(destination, data, 0644)
	if err != nil {
	}
	return err
}

//TruncateFile TruncateFile
func TruncateFile(path string, size int64) error {
	return os.Truncate(path, size)
}

//SeekFile SeekFile
// func SeekFile(path string, offset int64) error {
// 	file, err := os.OpenFile(path, os.O_RDWR, 0600)
// 	if err != nil {
// 		return err
// 	}

// 	defer file.Close()

// 	_, err = file.Seek(offset, 0)

// 	return err
// }

//FileInfo FileInfo
func FileInfo(path string) (os.FileInfo, error) {
	return os.Stat(path)
}

//FileSearch FileSearch
func (fileHelperServiceObject *FileHelperServiceObject) FileSearch(fileName, path string) ([]string, error) {
	fileHelperServiceObject.searchResult = []string{}
	fileHelperServiceObject.searchFileName = fileName
	searchDirectory, err := os.Open(path)
	if err != nil {
		return fileHelperServiceObject.searchResult, err
	}
	defer searchDirectory.Close()

	testFileInfo, _ := searchDirectory.Stat()
	if !testFileInfo.IsDir() {
		return fileHelperServiceObject.searchResult, err
	}

	err = filepath.Walk(path, fileHelperServiceObject.findFile)
	if err != nil {
		return fileHelperServiceObject.searchResult, err
	}

	return fileHelperServiceObject.searchResult, nil

}

func (fileHelperServiceObject *FileHelperServiceObject) findFile(path string, fileInfo os.FileInfo, err error) error {

	if err != nil {
		return err
	}

	// get absolute path of the folder that we are searching
	absolute, err := filepath.Abs(path)
	if err != nil {
		return err
	}

	if fileInfo.IsDir() {
		testDir, err := os.Open(absolute)
		if err != nil {
			return err
		}
		testDir.Close()
		return nil
	}

	matched, err := filepath.Match(fileHelperServiceObject.searchFileName, fileInfo.Name())
	if err != nil {
		return err
	}

	if matched {
		add := absolute
		fileHelperServiceObject.searchResult = append(fileHelperServiceObject.searchResult, add)
	}

	return nil
}

//FileAvailabilityCheck checks whether file is available at given location
func FileAvailabilityCheck(filePath string) bool {
	fileInfo, err := os.Stat(filePath)
	if fileInfo == nil && err != nil {
		return false
	}
	return true
}

// CleanPath
func CleanPath(path string) string {
	path = strings.Replace(path, "\n", "\\n", -1)
	path = strings.Replace(path, "\t", "\\t", -1)
	path = strings.Replace(path, "\r", "\\r", -1)
	path = strings.Replace(path, "\b", "\\b", -1)
	path = strings.Replace(path, "\a", "\\b", -1)
	path = strings.Replace(path, "\v", "\\b", -1)
	//path = strings.Replace(path, '\', '/', -1)

	normalizedPath := filepath.Clean(path)
	normalizedPath = filepath.ToSlash(normalizedPath)

	if strings.HasSuffix(path, string(filepath.Separator)) || strings.HasSuffix(path, "/") {
		normalizedPath = normalizedPath + "/"
	}

	return normalizedPath
}

//UnzipWithDefaultModifiedTime - unzips and sets access time and modified time to time set during zipping files
func UnzipWithDefaultModifiedTime(archive, target string) error {
	reader, err := zip.OpenReader(archive)
	if err != nil {
		logginghelper.LogError(err)
		return err
	}
	mkcDirError := os.MkdirAll(target, 0755)
	if mkcDirError != nil {
		logginghelper.LogError(mkcDirError)
		return mkcDirError
	}
	for _, file := range reader.File {
		filePath := filepath.Join(target, file.Name)
		parentPath := path.Dir(CleanPath(filePath))

		fileInfo := file.FileInfo()
		timeHeader := fileInfo.ModTime()

		if !FileAvailabilityCheck(parentPath) {
			CreateDirectoryRecursive(parentPath)
		}

		if fileInfo.IsDir() {
			os.MkdirAll(filePath, file.Mode())
			continue
		}

		fileReader, openError := file.Open()
		if openError != nil {
			logginghelper.LogError(openError)
			return openError
		}

		targetFile, targetOpenError := os.OpenFile(filePath, os.O_WRONLY|os.O_CREATE|os.O_TRUNC, file.Mode())
		if targetOpenError != nil {
			logginghelper.LogError(targetOpenError)
			return targetOpenError
		}

		_, copyError := io.Copy(targetFile, fileReader)
		if copyError != nil {
			logginghelper.LogError(copyError)
			return copyError
		}

		targetCloseError := targetFile.Close()
		if targetCloseError != nil {
			logginghelper.LogError(targetCloseError)
			return targetCloseError
		}

		fileCloseError := fileReader.Close()
		if fileCloseError != nil {
			logginghelper.LogError(fileCloseError)
			return fileCloseError
		}
		changeError := os.Chtimes(filePath, timeHeader.Local(), timeHeader.Local())
		if changeError != nil {
			logginghelper.LogError(changeError)
			return changeError
		}
	}
	closeError := reader.Close()
	if closeError != nil {
		logginghelper.LogError(closeError)
		return closeError
	}
	return nil
}
