package securityhelper

import (
	"bytes"
	"crypto/aes"
	"crypto/cipher"
	"crypto/md5"
	"encoding/base64"
	"errors"
	"os"
	"strconv"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/coreospackage/filehelper"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/coreospackage/logginghelper"

	OneOfOne "github.com/OneOfOne/xxhash"
)

var IV = "AAAAAAAAAAAAAAAA"

var securityKey = []byte{}

func SetSecurityConfig(secKey []byte, initializationVector string) {
	securityKey = secKey
	SetIV(initializationVector)
}

func SetIV(initializationVector string) {
	IV = initializationVector
}

//AESEncrypt Encrypts given text
func AESEncrypt(plainText, key []byte) ([]byte, error) {
	block, err := aes.NewCipher(key)
	if err != nil {
		return nil, err
	}

	padding := block.BlockSize() - len(plainText)%block.BlockSize()

	padtext := bytes.Repeat([]byte{byte(padding)}, padding)

	plainText = append(plainText, padtext...)

	cipherText := make([]byte, len(plainText))

	cbc := cipher.NewCBCEncrypter(block, []byte(IV))
	cbc.CryptBlocks(cipherText, plainText)

	encodedData := make([]byte, base64.StdEncoding.EncodedLen(len(plainText)))

	base64.StdEncoding.Encode(encodedData, cipherText)

	return encodedData, nil
}

//AESDecrypt Decrypts given cipher text
func AESDecrypt(encodedData, key []byte) ([]byte, error) {
	decodedData := make([]byte, base64.StdEncoding.DecodedLen(len(encodedData)))
	n, err := base64.StdEncoding.Decode(decodedData, encodedData)
	if err != nil {
		return nil, err
	}
	cipherText := decodedData[:n]

	block, err := aes.NewCipher(key)
	if err != nil {
		return nil, err
	}

	cbc := cipher.NewCBCDecrypter(block, []byte(IV))

	if len(cipherText)%cbc.BlockSize() != 0 {
		return nil, errors.New("crypto/cipher: input not full blocks")
	}

	cbc.CryptBlocks(cipherText, cipherText)

	length := len(cipherText)
	if length < 1 {
		logginghelper.LogError("length of cipherText is less than 1")
		return nil, errors.New("length of cipherText is less than 1")
	}

	unpadding := int(cipherText[length-1])

	if (length - unpadding) < 0 {
		return nil, errors.New("length of (length - unpadding) is less than 0")
	}

	return cipherText[:(length - unpadding)], nil
}

// GetHashChecksumOfByteArray Get Hash Check sum
func GetHashChecksumOfByteArray(byteArray []byte) (uint64, error) {
	h := OneOfOne.New64()
	_, err := h.Write(byteArray)
	if err != nil {
		return 0, err
	}
	return h.Sum64(), nil
}

//GetHashChecksumOfFile Get Hash Check sum
func GetHashChecksumOfFile(filePath string) (uint64, error) {
	byteArray, err := filehelper.ReadFile(filePath)
	if err != nil {
		return 0, err
	}
	hash, err := GetHashChecksumOfByteArray(byteArray)
	if err != nil {
		return 0, err
	}

	return hash, nil
}

// GetAttributeBasedHash GetAttributeBasedHash
func GetAttributeBasedHash(filePath string) (string, error) {
	fileInfo, statErr := os.Stat(filePath)
	if statErr != nil {
		logginghelper.LogError("error occured while getting stats of file : ", filePath, " : ", statErr)
		return "", statErr
	}
	fileModTime := fileInfo.ModTime()
	fileLength := fileInfo.Size()
	customFileHash := strconv.FormatInt(fileModTime.Unix(), 10) + strconv.FormatInt(fileLength, 10)
	return customFileHash, nil
}

// GetAttributeBasedHashBasedOnParameters - return custom hash based on file attributes
// func GetAttributeBasedHashByParameters(fileModTime string, fileSize string) (string, error) {
// 	modTime, _ := strconv.ParseInt(fileModTime, 10, 64)
// 	fSize, _ := strconv.ParseInt(fileSize, 10, 64)
// 	customFileHash := strconv.FormatInt((modTime + fSize), 10)
// 	return customFileHash, nil
// }

//GetAttributeBasedHashBasedOnParameters - return custom hash based on file attributes
func GetAttributeBasedHashByParameters(fileModTime string, fileSize string) (string, error) {
	modTime, parseError := strconv.ParseInt(fileModTime, 10, 64)
	if parseError != nil {
		logginghelper.LogError("error occured while parsing fileModTime : ", parseError)
		return "", parseError
	}
	cModTime := time.Unix(modTime, 0)
	date, time := TimeToMsDosTime(cModTime)
	newModTime := MsDosTimeToTime(date, time)
	fSize, parseFileSizeError := strconv.ParseInt(fileSize, 10, 64)
	if parseFileSizeError != nil {
		logginghelper.LogError("error occured while parsing fileModTime : ", parseFileSizeError)
		return "", parseFileSizeError
	}
	// customFileHash := strconv.FormatInt((newModTime + fSize), 10)
	customFileHash := strconv.FormatInt(newModTime.Unix(), 10) + strconv.FormatInt(fSize, 10)
	return customFileHash, nil
}

//AESEncryptDefault method to encrypt with default config (securityKey)
func AESEncryptDefault(plainText []byte) ([]byte, error) {
	compressedText, compressionError := Compress(plainText)
	if compressionError != nil {
		return plainText, compressionError
	}
	return AESEncrypt(compressedText, securityKey)
}

// AESEncryptDefault method to decrypt with default config (securityKey)
func AESDecryptDefault(encodedData []byte) ([]byte, error) {

	compressedText, decryptionError := AESDecrypt(encodedData, securityKey)
	if decryptionError != nil {
		return encodedData, decryptionError
	}
	return Decompress(compressedText)
}

// TODO: implement another hashing algo instead of MD5
// Get128BitHash returns 128 bit hash
func Get128BitHash(data []byte) [16]byte {
	return md5.Sum(data)
}

// GetCespareXXHash128 returns 128 bit  xxhash, which is calculated using cespare
// func GetCespareXXHash128(data string) []byte {
// 	// x := "MKCLSecurity$#@!"
// 	sum := cespare.Sum64String(data)
// 	// x := byte(sum)
// 	// fmt.Println(">> ", x)
// 	// hash := (*[16]byte)(unsafe.Pointer(&sum))[:]
// 	hashString := strconv.FormatUint(sum, 6)
// 	return []byte(hashString)
// }

// GetAttributeBasedHashWithRoundedModTime - rounds of the mod time similar to rounding done during zipping
func GetAttributeBasedHashWithRoundedModTime(filePath string) (string, error) {
	fileInfo, statErr := os.Stat(filePath)
	if statErr != nil {
		logginghelper.LogError("error occured while getting stats of file : ", filePath, " : ", statErr)
		return "", statErr
	}
	date, time := TimeToMsDosTime(fileInfo.ModTime())
	newModTime := MsDosTimeToTime(date, time)
	fileLength := fileInfo.Size()
	customFileHash := strconv.FormatInt(newModTime.Unix(), 10) + strconv.FormatInt(fileLength, 10)
	return customFileHash, nil
}

func TimeToMsDosTime(t time.Time) (fDate uint16, fTime uint16) {
	fDate = uint16(t.Day() + int(t.Month())<<5 + (t.Year()-1980)<<9)
	fTime = uint16(t.Second()/2 + t.Minute()<<5 + t.Hour()<<11)
	return
}

func MsDosTimeToTime(dosDate, dosTime uint16) time.Time {
	return time.Date(
		// date bits 0-4: day of month; 5-8: month; 9-15: years since 1980
		int(dosDate>>9+1980),
		time.Month(dosDate>>5&0xf),
		int(dosDate&0x1f),

		// time bits 0-4: second/2; 5-10: minute; 11-15: hour
		int(dosTime>>11),
		int(dosTime>>5&0x3f),
		int(dosTime&0x1f*2),
		0, // nanoseconds

		time.UTC,
	)
}
